
-- CotW map functions

local M = {}

-- UTILITY FUNCTIONS --

local function smooth(map, range)
	local smoothed = {}
	for x = 1, #map do
		smoothed[x] = {}
		for y = 1, #map[x] do
			local sum, count = 0, 0
			for i = math.max(x - range, 1), math.min(x + range, #map) do
				for j = math.max(y - range, 1), math.min(y + range, #map[x]) do
					local value = 2^((range*2) - math.abs(x-i) - math.abs(y-j))
					sum = sum + map[i][j] * value
					count = count + value
				end
			end
			smoothed[x][y] = sum/count
		end
	end
	return smoothed
end

local function getDistance(x1, y1, x2, y2)
	return math.abs(x1 - x2) + math.abs(y1 - y2)
end

-- MAP CREATION FUNCTIONS --

function cellularAutomata(mapWidth, mapLength)

	-- seed map
	local solidChance, map = 0.55, {}
	for x = 1, mapWidth do
		map[x] = {}
		for y = 1, mapLength do
			map[x][y] = love.math.random() < solidChance and 1 or 0
		end
	end

	-- cellular automata process (using 4 generations of the 4-5 rule)
	for gen = 1, 4 do
		for x = 1, mapWidth do
			for y = 1, mapLength do
				-- count current neighbors
				local neighborCount = 0
				for i = math.max(x-1,1), math.min(x+1,mapWidth) do
					for j = math.max(y-1,1), math.min(y+1,mapLength) do
						neighborCount = neighborCount + map[i][j]
					end
				end
				map[x][y] = neighborCount > 4 and 1 or 0
			end
		end
	end

	return map
end

function M.createNewWorld(mapWidth, mapLength)

	-- Initialize world creation
	--					 Ocean River Desrt Grass Trees Swamp Rocks Mount
	local terrainCost = {7.7,  3.6,  1.5,  1.0,  1.3,  2.0,  1.0,  7.5}
	local terrainVal  = {12,   6,    1,    3,    6,    2,    13,   9}
	local map = {w=mapWidth, h=mapLength, terDefault=terOcean, roadCostDirt=0.2, roadCostStone=0.1, avgCost=0.2}
	map.roadCost = math.max(map.roadCostDirt, map.roadCostStone)

	-- MAP TERRAIN

	-- Create heightmap
	local mapHeight = diamondsquare.create(mapWidth, mapLength)

	-- Apply cellular automata terrain mask
	local mapMask = cellularAutomata(mapWidth, mapLength)
	for x = 1, mapWidth do
		for y = 1, mapLength do
			mapHeight[x][y] = mapHeight[x][y] * mapMask[x][y]
		end
	end
	mapMask = nil

	-- Smooth heightmap
	mapHeight = smooth(mapHeight, 4)
	mapHeight = smooth(mapHeight, 2)
	mapHeight = smooth(mapHeight, 1)

	-- Create rainfall map
	local mapRain = diamondsquare.create(mapWidth, mapLength, 1)
	mapRain = smooth(mapRain, 3)

	-- Create rivers
	local HEIGHT = {Water=0.2, Ground=0.65, Spring=0.63, NewRiver=1.0}
	local RAIN = {Desrt=0.2, Grass=0.6, Trees=0.85, Spring=0.93}
	for x = 1, mapWidth do
		for y = 1, mapLength do
			-- rivers start at high-altitude rainy places
			if mapHeight[x][y] >= HEIGHT.Spring and mapRain[x][y] > RAIN.Spring then
				local river = {x=x, y=y, z, dx, dy}
				river[1] = {x=x, y=y}
				repeat
					-- run the river through lowest neighbor until water or map edge
					mapHeight[river.x][river.y] = HEIGHT.NewRiver
					river.z = HEIGHT.NewRiver -- this is different than .Water so the river doesn't end when it touches itself
					if river.x > 1 and mapHeight[river.x-1][river.y] < river.z then
						river.dx = -1; river.dy = 0; river.z = mapHeight[river.x-1][river.y] end
					if river.y > 1 and mapHeight[river.x][river.y-1] < river.z then
						river.dx = 0; river.dy = -1; river.z = mapHeight[river.x][river.y-1] end
					if river.x < mapWidth and mapHeight[river.x+1][river.y] < river.z then
						river.dx = 1; river.dy = 0; river.z = mapHeight[river.x+1][river.y] end
					if river.y < mapLength and mapHeight[river.x][river.y+1] < river.z then
						river.dx = 0; river.dy = 1; river.z = mapHeight[river.x][river.y+1] end
					if river.dx == nil then -- if the river gets trapped, make deep water
						mapHeight[river.x][river.y] = 0
					else
						river.x = river.x + river.dx
						river.y = river.y + river.dy
						river[#river+1] = {x = river.x, y = river.y}
					end
				until mapHeight[river.x][river.y] <= HEIGHT.Water or river.x == 1 or river.x == mapWidth or river.y == 1 or river.y == mapLength
				if mapHeight[river.x][river.y] <= HEIGHT.Water then river[#river] = nil end -- don't overwrite existing water
				-- apply the defined river
				for i = 1, #river do
					mapHeight[river[i].x][river[i].y] = HEIGHT.Water
				end
			end
		end
	end

	-- Create Terrain
	for x = 1, mapWidth do
		map[x] = {}
		for y = 1, mapLength do
			local height, rain, floor, wall = mapHeight[x][y], mapRain[x][y], true, false
			if     height  < HEIGHT.Water then terType = terOcean; floor = false
			elseif height == HEIGHT.Water then terType = terRiver; floor = false
			elseif height  < HEIGHT.Ground then
				if     rain < RAIN.Desrt then terType = terDesrt
				elseif rain < RAIN.Grass then terType = terGrass
				elseif rain < RAIN.Trees then terType = terTrees -- *** Use more kinds of trees ??
				else   terType = terSwamp -- *** allow latitude to influence swamp cutoff ??
				end
			else
				terType = terMount; wall = true
			end
			map[x][y] = {terrain=terType, pathCost=terrainCost[terType], color=terType, hasFloor=floor, hasWall=wall}
		end
	end
	mapHeight = nil
	mapRain = nil


	-- WORLD MAP LOCATIONS

	-- Initial location nodes
	local nodeByCell = {}
	for x = 1, mapWidth do
		nodeByCell[x] = {}
	end
	local count = math.floor((mapWidth*mapLength)^0.3)
	countLocations = tostring(count) -- DEBUG ************
	local nodeByID = {}
	for n = 1, count do
		local x, y
		repeat
			x = love.math.random(1, mapWidth)
			y = love.math.random(1, mapLength)
		until not nodeByCell[x][y]
		nodeByID[n] = {x=x, y=y, size=0, value=0}
		nodeByCell[x][y] = n -- to prevent picking the same coords twice
	end

	-- Calculate Voroni regions
	for x = 1, mapWidth do
		for y = 1, mapLength do
			if map[x][y].hasFloor and not map[x][y].hasWall then
				local ID, dist = 0, mapWidth + mapLength
				for n = 1, #nodeByID do
					if getDistance(x, y, nodeByID[n].x, nodeByID[n].y) < dist then
						ID = n
						dist = getDistance(x, y, nodeByID[n].x, nodeByID[n].y)
					end
				end
				nodeByCell[x][y] = ID
			end
		end
	end

	-- Apply Lloyd relaxation
	for n = 1, #nodeByID do
		local sum = {x=0, y=0, count=0}
		for x = 1, mapWidth do
			for y = 1, mapLength do
				if nodeByCell[x][y] == n then
					sum.count = sum.count + 1
					sum.x = sum.x + x
					sum.y = sum.y + y
				end
			end
		end
		nodeByID[n].x = math.floor(sum.x/sum.count)
		nodeByID[n].y = math.floor(sum.y/sum.count)
	end

	-- Update Voroni Regions
	for x = 1, mapWidth do
		for y = 1, mapLength do
			if map[x][y].hasFloor and not map[x][y].hasWall then
				local ID, dist = 0, mapWidth + mapLength
				for n = 1, #nodeByID do
					if getDistance(x, y, nodeByID[n].x, nodeByID[n].y) < dist then
						ID = n
						dist = getDistance(x, y, nodeByID[n].x, nodeByID[n].y)
					end
				end
				nodeByCell[x][y] = ID
				nodeByID[ID].size = nodeByID[ID].size + 1 -- this is where we start tracking region size
			end
		end
	end

	-- Normalize region sizes
	local max, min = 0, mapWidth*mapLength
	for n = 1, #nodeByID do
		max = nodeByID[n].size > max and nodeByID[n].size or max
		min = nodeByID[n].size < min and nodeByID[n].size or min
	end
	for n = 1, #nodeByID do
		nodeByID[n].size = (nodeByID[n].size - min) / (max-min)
	end

	-- Place Location at highest-value location in each Region
	for n = #nodeByID, 1, -1 do -- go in reverse order so nothing skips or crashes when we drop a node and increment the counter
		local range = 3
		local Location = {x=0,y=0,value=0}
		for x = 1, mapWidth do
			for y = 1, mapLength do
				if nodeByCell[x][y] == n then
					local value = 0
					for i = math.max(x-range,1), math.min(x+range,mapWidth) do
						for j = math.max(y-range,1), math.min(y+range,mapLength) do
							value = value + terrainVal[map[i][j].terrain]
							if nodeByCell[i][j] == n then value = value + terrainVal[map[i][j].terrain] end -- double points for cells in your OWN region
						end
					end
					if value > Location.value and map[x][y].hasFloor and not map[x][y].hasWall then
						Location.x = x
						Location.y = y
						Location.value = value
					end
				end
			end
		end
		-- remove any invalid Locations
		if Location.value == 0 then
			table.remove(nodeByID, n)
		else
			nodeByID[n].value = Location.value
			nodeByID[n].x = Location.x
			nodeByID[n].y = Location.y
			map[Location.x][Location.y].pathCost = map.roadCostStone
		end
	end
	nodeByCell = nil

	-- Calculate average node value
	nodeByID.avgVal = 0
	for n = 1, #nodeByID do
		nodeByID.avgVal = nodeByID.avgVal + nodeByID[n].value/#nodeByID
	end

	-- Set Location type
	-- HACK *** Maybe just set alignment here, and use the number of connected nodes to determine "magnitude" ?
	for n = 1, #nodeByID do
		if nodeByID[n].value > nodeByID.avgVal then
			nodeByID[n].align = true -- hamlets, towns, cities
			if     nodeByID[n].size < 0.33 then nodeByID[n].feature = ftrHmlt
			elseif nodeByID[n].size < 0.67 then nodeByID[n].feature = ftrTown
			else                                nodeByID[n].feature = ftrCity
			end
			player.x = nodeByID[n].x -- start player at a good Location
			player.y = nodeByID[n].y
		else
			nodeByID[n].align = false -- caves, ruins, dungeons
			if     nodeByID[n].size < 0.33 then nodeByID[n].feature = ftrCave
			elseif nodeByID[n].size < 0.67 then nodeByID[n].feature = ftrRuin
			else                                nodeByID[n].feature = ftrDngn
			end
		end
	end


	-- LOCATION CONNECTING (ROAD BUILDING ETC)

	-- Calculate Manhattan distances between Locations
	local nodePaths = {}
	for n1 = 1, #nodeByID - 1 do
		for n2 = n1 + 1, #nodeByID do
			local dist = math.abs(nodeByID[n1].x - nodeByID[n2].x) + math.abs(nodeByID[n1].y - nodeByID[n2].y)
			if dist < 100 then nodePaths[#nodePaths + 1] = {n1=n1, n2=n2, x1=nodeByID[n1].x, y1=nodeByID[n1].y, x2=nodeByID[n2].x, y2=nodeByID[n2].y, dist=dist} end
		end
	end

	-- Sort by Ascending distance (to build shortest-first)
	for i = 1, #nodePaths - 1 do
		for j = i+1, #nodePaths do
			if nodePaths[i].dist > nodePaths[j].dist then
				local tmp = nodePaths[j]
				nodePaths[j] = nodePaths[i]
				nodePaths[i] = tmp
			end
		end
	end

	-- Determine paths to build, starting with path #1
	local nodePathsToBuild, nodesIncluded, n = {1}, {}
	nodesIncluded[nodePaths[1].n1] = true
	nodesIncluded[nodePaths[1].n2] = true
	-- select unconnected nodes until they're all linked
	repeat
		local i, n1, n2 = 0
		repeat -- choose next unconnected node
			i = i + 1
			if nodePaths[i] then n1, n2 = nodePaths[i].n1, nodePaths[i].n2 end
		until i > #nodePaths or nodesIncluded[n1] ~= nodesIncluded[n2]
		if nodePaths[i] then
			nodesIncluded[n1], nodesIncluded[n2] = true, true
			nodePathsToBuild[#nodePathsToBuild+1] = i
		end
	until i > #nodePaths

	-- Build paths
	local paths = {}
	for i = 1, #nodePathsToBuild do
		local n1, n2 = nodePaths[nodePathsToBuild[i]].n1, nodePaths[nodePathsToBuild[i]].n2
		paths[i] = ai.pathfind(map, nodeByID[n1].x, nodeByID[n1].y, nodeByID[n2].x, nodeByID[n2].y)
		-- determine what type of road to apply
		local roadCost
		if nodeByID[n1].align and nodeByID[n2].align then -- only Good locations generate roads
			if nodeByID[n1].feature == ftrCity or nodeByID[n2].feature == ftrCity then
				roadCost = map.roadCostStone
			elseif nodeByID[n1].feature == ftrTown or nodeByID[n2].feature == ftrTown then
				roadCost = map.roadCostDirt
			end
		end
		-- apply doors to map
		local door
		for j = 1, #paths[i] do
			local x, y = paths[i][j].x, paths[i][j].y
			-- door end point?
			if door and map[x][y].hasFloor and not map[x][y].hasWall then
				local x1, y1, x2, y2 = door.x, door.y, paths[i][j-1].x, paths[i][j-1].y
				map[x1][y1].terrain = door.terrain
				map[x2][y2].terrain = door.terrain
				map[x1][y1].hasFloor = true
				map[x2][y2].hasFloor = true
				map[x1][y1].hasWall = false
				map[x2][y2].hasWall = false
				if getDistance(x1, y1, x2, y2) > 2 then
					-- apply door endpoints
					map[x1][y1].feature = door.feature
					map[x2][y2].feature = door.feature
					map[x1][y1].door = {map=1, x=x2, y=y2}
					map[x2][y2].door = {map=1, x=x1, y=y1}
				end
				door = nil
			-- begin Tunnel?
			elseif not door and map[x][y].hasWall then
				door = {x=x, y=y, feature=ftrTunl, terrain=terRocks}
			-- begin Boat?
			elseif not door and not map[x][y].hasFloor then
				door = {x=x, y=y, feature=ftrBoat, terrain=map[x][y].terrain}
			end
			-- apply path
			if not door and roadCost then
				for j = 1, #paths[i] do
					local x, y = paths[i][j].x, paths[i][j].y
					if roadCost < map[x][y].pathCost then map[x][y].pathCost = roadCost end
				end
			end
		end
	end

	-- Apply path graphics
	for x = 1, mapWidth do
		for y = 1, mapLength do
			if not map[x][y].feature and map[x][y].pathCost <= map.roadCost and map[x][y].hasFloor and not map[x][y].hasWall then
				local roadShape = 0
				if map[x][y].pathCost == map.roadCostStone then roadShape = 15 end
				if x > 1 and map[x-1][y].pathCost <= map.roadCost then roadShape = roadShape + 1 end
				if y > 1 and map[x][y-1].pathCost <= map.roadCost then roadShape = roadShape + 2 end
				if x < mapWidth  and map[x+1][y].pathCost <= map.roadCost then roadShape = roadShape + 4 end
				if y < mapLength and map[x][y+1].pathCost <= map.roadCost then roadShape = roadShape + 8 end
				map[x][y].feature = roadShape
				map[x][y].color = terRocks
				local terrain = map[x][y].terrain
				if terrain == terOcean or terrain == terRiver then
					map[x][y].hasFloor = true
				elseif terrain == terTrees then
					map[x][y].terrain = terGrass
				elseif terrain == terMount then
					map[x][y].terrain = terRocks
					map[x][y].hasWall = false
				end
			end
		end
	end

	-- Set Location graphics
	for n = 1, #nodeByID do
		local x, y = nodeByID[n].x, nodeByID[n].y
		map[x][y].feature = nodeByID[n].feature
	end


	-- LOCATION GENERATION
	-- ***
	-- City gen
	-- ...
	-- https://procworld.blogspot.com/2011/07/city-lots.html?m=1
	-- ...


	return {map} -- HACK *** Return all maps here, with world map being [1] etc
end

return M
